import jax
import jax.numpy as jnp
import equinox as eqx
import equinox.internal as eqxi

from tests.test_jax import run_and_compare


def test_conv_1d():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.Conv1d

        def __init__(self, key):
            self.conv = eqx.nn.Conv1d(in_channels=2, out_channels=5, kernel_size=3, key=key)

        def __call__(self, x):
            return self.conv(x)

    input_spec = (jnp.zeros((2, 30)), )
    model = SimpleConv(jax.random.PRNGKey(0))
    run_and_compare(eqxi.finalise_fn(model), input_spec)


def test_conv_transpose_1d():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.ConvTranspose1d

        def __init__(self, key):
            self.conv = eqx.nn.ConvTranspose1d(in_channels=5, out_channels=2, kernel_size=3, key=key)

        def __call__(self, x):
            return self.conv(x)

    input_spec = (jnp.zeros((5, 30)), )
    model = SimpleConv(jax.random.PRNGKey(0))
    run_and_compare(eqxi.finalise_fn(model), input_spec)


def test_conv_2d():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.Conv2d

        def __init__(self, key):
            self.conv = eqx.nn.Conv2d(in_channels=2, out_channels=5, kernel_size=(3, 4), key=key)

        def __call__(self, x):
            return self.conv(x)

    input_spec = (jnp.zeros((2, 40, 30)), )
    model = SimpleConv(jax.random.PRNGKey(0))
    run_and_compare(eqxi.finalise_fn(model), input_spec)


def test_conv_transpose_2d():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.ConvTranspose2d

        def __init__(self, key):
            self.conv = eqx.nn.ConvTranspose2d(in_channels=5, out_channels=2, kernel_size=(3, 4), key=key)

        def __call__(self, x):
            return self.conv(x)

    input_spec = (jnp.zeros((5, 40, 30)), )
    model = SimpleConv(jax.random.PRNGKey(0))
    run_and_compare(eqxi.finalise_fn(model), input_spec)


def test_conv_3d():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.Conv3d

        def __init__(self, key):
            self.conv = eqx.nn.Conv3d(in_channels=2, out_channels=5, kernel_size=(3, 4, 2), key=key)

        def __call__(self, x):
            return self.conv(x)

    input_spec = (jnp.zeros((2, 40, 30, 15)), )
    model = SimpleConv(jax.random.PRNGKey(0))
    run_and_compare(eqxi.finalise_fn(model), input_spec)


def test_conv_transpose_3d():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.ConvTranspose3d

        def __init__(self, key):
            self.conv = eqx.nn.ConvTranspose3d(in_channels=5, out_channels=2, kernel_size=(3, 4, 2), key=key)

        def __call__(self, x):
            return self.conv(x)

    input_spec = (jnp.zeros((5, 40, 30, 15)), )
    model = SimpleConv(jax.random.PRNGKey(0))
    run_and_compare(eqxi.finalise_fn(model), input_spec)


def test_odd_batch_dimension():
    class SimpleConv(eqx.Module):
        conv: eqx.nn.Conv1d

        def __init__(self, key):
            self.conv = eqx.nn.Conv1d(in_channels=2, out_channels=5, kernel_size=3, key=key)

        def __call__(self, x):
            return self.conv(x)

    model = SimpleConv(jax.random.PRNGKey(0))
    batched_model = jax.vmap(model, axis_name="batch", in_axes=2, out_axes=2)

    input_spec = (jnp.zeros((2, 30, 5)), )
    run_and_compare(eqxi.finalise_fn(batched_model), input_spec)



from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil.var import Var

from dataclasses import dataclass
from typing import List
from functools import reduce
import itertools


@dataclass
class ResolvedSliceSpec:
    start_indices: List[int] | Var
    end_indices: List[int] | Var
    strides: List[int]
    shape: List[int]


def index_by_slices(tensor, slice_spec):
    tensor = fix_scalar_tensor(tensor)
    resolved_slices = _resolve_slice_spec(tensor, slice_spec)

    return mb.slice_by_index(
        x=tensor,
        begin=resolved_slices.start_indices,
        end=resolved_slices.end_indices,
        stride=resolved_slices.strides
    )


def update_tensor_by_slice(tensor, slice_spec, value):
    tensor = fix_scalar_tensor(tensor)
    resolved_slices = _resolve_slice_spec(tensor, slice_spec)

    value = mb.reshape(x=value, shape=resolved_slices.shape)
    return mb.slice_update(
        x=tensor,
        update=value,
        begin=resolved_slices.start_indices,
        end=resolved_slices.end_indices,
        stride=resolved_slices.strides
    )


def fix_scalar_tensor(tensor):
    """
    From a numpy scalar type, CoreML will create a rank 0 tensor, which it will
    later struggle to do operations on. We will re-shape it to a rank 1 tensor
    with dimension 1.
    """
    if len(tensor.shape) == 0:
        tensor = mb.reshape(x=tensor, shape=(1,))
    return tensor


def _flatten_list(lst):
    flat_list = []
    for item in lst:
        if isinstance(item, (list, tuple)):
            flat_list += _flatten_list(item)
        else:
            flat_list.append(item)
    return flat_list


def _count_dimensions(slice_spec):
    dim_count = 0
    for spec in slice_spec:
        if isinstance(spec, Var):
            if spec.rank != 1:
                raise ValueError("The Var spec must have rank 1!")
            dim_count += spec.shape[0]
        elif isinstance(spec, type(Ellipsis)):
            raise ValueError("Can not count dimensions for slice spec containing Ellipsis")
        else:
            dim_count += 1
    return dim_count


def _resolve_slice_spec(tensor, slice_spec) -> ResolvedSliceSpec:
    start_indices = []
    end_indices = []
    strides = []
    shape = []

    # We allow the slice_spec to have nested lists. In that case we flatten it
    slice_spec = _flatten_list(slice_spec)
    if len(slice_spec) == 0:
        # Special case for scalar
        slice_spec = [slice(None)]

    tensor_rank = len(tensor.shape)
    contains_var_type = False
    dim_counter = 0
    for i, spec in enumerate(slice_spec):
        if isinstance(spec, type(slice(None))):
            start_indices.append(spec.start or 0)
            end_indices.append(spec.stop or tensor.shape[dim_counter])
            strides.append(spec.step or 1)
            shape.append(end_indices[-1] - start_indices[-1] // strides[-1])
            dim_counter += 1
        elif isinstance(spec, type(Ellipsis)):
            if any([isinstance(s, type(Ellipsis)) for s in slice_spec[i+1:]]):
                raise ValueError("Only supports one ellipsis when indexing")

            dims_before = dim_counter
            dims_after = _count_dimensions(slice_spec[i+1:])
            num_ellipsis_dims = tensor_rank - (dims_before + dims_after)

            ellipsis_starts = [0] * num_ellipsis_dims
            ellipsis_ends = [tensor.shape[dim] for dim in range(dim_counter, dim_counter + num_ellipsis_dims)]
            ellipsis_strides = [1] * num_ellipsis_dims
            ellipsis_shape = [
                (end - start) // stride for start, end, stride
                in zip(ellipsis_starts, ellipsis_ends, ellipsis_strides)
            ]

            start_indices += ellipsis_starts
            end_indices += ellipsis_ends
            strides += ellipsis_strides
            shape += ellipsis_shape

            dim_counter += num_ellipsis_dims
        elif isinstance(spec, Var):
            if spec.rank != 1:
                raise ValueError("The Var spec must have rank 1!")
            contains_var_type = True
            start_indices.append(spec)
            end_indices.append(mb.add(x=spec, y=1))
            strides += [1] * spec.shape[0]
            shape += [1] * spec.shape[0]
            dim_counter += spec.shape[0]
        else:
            # Assume it is an integer index
            idx = int(spec)
            start_indices.append(idx)
            end_indices.append(idx + 1)
            strides.append(1)
            shape.append(1)
            dim_counter += 1

    # If slice_spec contained any Var types, we will need to concatenate the full result
    # to be one big Var type
    if contains_var_type:
        def partition_list(lst):
            parts = [[]]
            for element in lst:
                if isinstance(element, Var):
                    if len(parts[-1]) == 0:
                        parts.pop()
                    parts.append(element)
                    parts.append([])
                else:
                    parts[-1].append(element)
            if len(parts[-1]) == 0:
                # The last partition may be empty, if so we skip it
                return parts[:-1]
            return parts

        def concat_to_var(lst):
            parts = partition_list(lst)
            return mb.concat(values=parts, axis=0)

        start_indices = concat_to_var(start_indices)
        end_indices = concat_to_var(end_indices)

    if len(strides) != tensor_rank or len(shape) != tensor_rank:
        raise ValueError("Slice does not line up!")

    return ResolvedSliceSpec(
        start_indices=start_indices,
        end_indices=end_indices,
        strides=strides,
        shape=shape,
    )


def iterate_indexes_in_shapes(func, shapes: List, init_values: List, unroll_limit: int = 25):
    """
    Given a list of `shapes`, fx [(3, 2, 3), (5, 2, 3)] this method will iterate
    the product of all valid indexes into the given shapes.
    The function `func: Idx1, Idx2, ..., Idxn, Acc1, ..., Acck -> Res1, ..., Resk` is expected to given the
    list of indexes and the accumulated result so far, to update the result based
    on the index.
    The init_values is a list of [InitVal1, ..., InitValk], and the function `func`
    must return a list of `k` values [Res1, ..., Resk].
    The indexes, Idx1, ..., Idn, may be either a mil mb.Var type of a python
    tuple depending on if the loop is unrolled or not. `func` is expected to be
    able to handle this.

    If the total number of traversed indexes is <`unroll_limit`, the loop will be
    fully un-rolled into MIL instructions. Otherwise it will be constructed as
    a dynamic while loop executed at runtime.
    """
    shapes_elements = [reduce(lambda a, b: a * b, shape, 1) for shape in shapes]
    total_iterations = reduce(lambda a, b: a * b, shapes_elements, 1)

    results = init_values
    if total_iterations <= unroll_limit:
        # Fully unroll the loop
        ranges = [itertools.product(*[range(dim) for dim in shape]) for shape in shapes]
        for indexes in itertools.product(*ranges):
            results = func(*indexes, *results)
    else:
        # Dynamically compute the loop
        def suffix_product(lst):
            res = []
            acc = 1
            for i in reversed(lst):
                res.append(acc)
                acc *= i
            return list(reversed(res))
        integer_index_strides = suffix_product(shapes_elements)
        index_strides = [suffix_product(shape) for shape in shapes]

        # Attempt at looping over result indexes without fully unrolling
        def cond(i, *acc):
            return mb.less(x=i, y=total_iterations)

        def body(i, *acc):
            # Split out the index i to an integer index into the individual shapes.
            integer_indexes = [
                mb.mod(x=mb.floor_div(x=i, y=stride), y=elements)
                for stride, elements in zip(integer_index_strides, shapes_elements)
            ]
            # Map the integer index in the shapes to an actual shaped index
            indexes = [
                mb.concat(values=[
                    mb.mod(x=mb.floor_div(x=idx, y=stride), y=dim) for stride, dim in zip(strides, shape)
                ], axis=0)
                if len(shape) > 0 else []
                for idx, strides, shape in zip(integer_indexes, index_strides, shapes)
            ]

            results = func(*indexes, *acc)
            return [mb.add(x=i, y=1)] + results

        # TODO: Fix this in a nicer way!!!
        fixed_init_values = []
        for init_value in init_values:
            if len(init_value.shape) == 0:
                fixed_init_values.append(mb.reshape(x=init_value, shape=(1,)))
            else:
                fixed_init_values.append(init_value)

        results = mb.while_loop(_cond=cond, _body=body, loop_vars=[0] + fixed_init_values)[1:]  # Skip the counter

    return results


def inverse_permutation(perm):
    """
    Given a permutation `perm`, compute the inverse of the permutation
    """
    inv = [0] * len(perm)
    for i, j in enumerate(perm):
        inv[j] = i
    return inv



from flax import nnx
import jax.numpy as jnp

from typing import List

from .flax_xlstm import xLSTM


class ResidualConv(nnx.Module):
    scale_conv: nnx.Conv
    conv: nnx.Conv
    normalization_1: nnx.Module
    normalization_2: nnx.Module
    normalization_3: nnx.Module
    shortcut: nnx.Conv

    def __init__(self, in_channels: int, out_channels: int, rngs: nnx.Rngs, stride: int = 2):
        conv_type = nnx.Conv if in_channels <= out_channels else nnx.ConvTranspose

        kernel_size = 4
        self.scale_conv = conv_type(
            in_features=in_channels,
            out_features=out_channels,
            kernel_size=(kernel_size,),
            strides=(stride,),
            rngs=rngs
        )
        self.conv = nnx.Conv(
            in_features=out_channels,
            out_features=out_channels,
            kernel_size=kernel_size,
            rngs=rngs,
        )

        self.normalization_1 = nnx.BatchNorm(num_features=out_channels, rngs=rngs)
        self.normalization_2 = nnx.BatchNorm(num_features=out_channels, rngs=rngs)

        self.shortcut = conv_type(
            in_features=in_channels,
            out_features=out_channels,
            kernel_size=(stride,),
            strides=(stride,),
            rngs=rngs
        )

    def __call__(self, x):
        out = self.scale_conv(x)
        out = self.normalization_1(out)
        out = nnx.silu(out)

        out = self.conv(out)
        out = nnx.silu(out)

        # Residual
        out = out + self.shortcut(x)
        out = self.normalization_2(out)

        return out


class Encoder(nnx.Module):
    cnn_layers: List[ResidualConv]
    normalization: nnx.Module

    def __init__(self, num_layers: int, rngs: nnx.Rngs):
        self.cnn_layers = []

        for i in range(num_layers):
            in_channels = (2 ** i)
            out_channels = 2 ** (i + 1)

            self.cnn_layers.append(ResidualConv(
                in_channels=in_channels,
                out_channels=out_channels,
                rngs=rngs,
            ))

        last_layer_features = 2 ** num_layers
        self.normalization = nnx.BatchNorm(num_features=last_layer_features, rngs=rngs)

    def __call__(self, x):
        out = x
        skip_connections = []
        for layer in self.cnn_layers:
            out = layer(out)
            skip_connections.append(out)

        out = self.normalization(out)
        out = nnx.tanh(out)

        return out, skip_connections


class Decoder(nnx.Module):
    cnn_layers: List[ResidualConv]
    residual_norm_layers: List[nnx.Module]
    output_polling: nnx.Conv

    def __init__(self, num_layers: int, rngs: nnx.Rngs):
        self.cnn_layers = []
        self.residual_norm_layers = []

        input_features = 2 ** num_layers
        for i in range(num_layers):
            # Times two to handle residual connections
            in_channels = 2 * (input_features // (2 ** i))
            out_channels = input_features // (2 ** (i + 1))

            self.residual_norm_layers.append(nnx.BatchNorm(in_channels, rngs=rngs))
            self.cnn_layers.append(ResidualConv(
                in_channels=in_channels,
                out_channels=out_channels,
                rngs=rngs,
            ))

        last_layer_features = input_features // (2 ** num_layers)
        self.output_polling = nnx.Conv(
            in_features=last_layer_features,
            out_features=1,
            kernel_size=3,
            rngs=rngs,
        )

    def __call__(self, x, skip_connections):
        skip_connections = list(reversed(skip_connections))

        out = x
        for i, (cnn_layer, residual_norm) in enumerate(zip(self.cnn_layers, self.residual_norm_layers)):
            residual = skip_connections[i]
            out = residual_norm(jnp.concatenate([out, residual], axis=-1))
            out = cnn_layer(out)

        out = self.output_polling(out)
        return out


class UNet(nnx.Module):
    encoder: Encoder
    decoder: Decoder

    def __init__(self, num_layers: int, rngs: nnx.Rngs):
        self.audio_encoding = Encoder(num_layers=num_layers, rngs=rngs)
        self.audio_decoding = Decoder(num_layers=num_layers, rngs=rngs)

    def __call__(self, x):
        def compress_dynamic_range(samples):
            mu = jnp.array(255.0, dtype=jnp.float16)
            return jnp.sign(samples) * jnp.log1p(mu * jnp.abs(samples)) / jnp.log1p(mu)
        x = compress_dynamic_range(x)

        hidden, skip_connections = self.audio_encoding(x)
        out = self.audio_decoding(hidden, skip_connections)

        return out


class UNetWithXlstm(nnx.Module):
    encoder: Encoder
    decoder: Decoder
    xlstm: xLSTM

    def __init__(self, num_conv_layers: int, rngs: nnx.Rngs):
        self.hidden_size = 2 ** num_conv_layers

        self.audio_encoding = Encoder(num_layers=num_conv_layers, rngs=rngs)
        self.audio_decoding = Decoder(num_layers=num_conv_layers, rngs=rngs)
        self.xlstm = xLSTM(hidden_size=self.hidden_size, num_heads=1, num_layers=1, rngs=rngs)

    def __call__(self, carry, x):
        if x.shape[1] != self.hidden_size or x.shape[2] != 1:
            raise ValueError("The input must be exactly squeezed to length 1 with x.shape[1] output features")

        def compress_dynamic_range(samples):
            mu = jnp.array(255.0, dtype=jnp.float16)
            return jnp.sign(samples) * jnp.log1p(mu * jnp.abs(samples)) / jnp.log1p(mu)
        x = compress_dynamic_range(x)

        hidden, skip_connections = self.audio_encoding(x)
        hidden = jnp.squeeze(hidden, axis=1)
        carry, hidden = self.xlstm(carry, hidden)
        hidden = jnp.expand_dims(hidden, axis=1)
        out = self.audio_decoding(hidden, skip_connections)

        return carry, out

    def init_carry(self, batch_size: int, rngs: nnx.Rngs):
        return self.xlstm.init_carry(batch_size, rngs)



from coremltools import _logger as logger
from coremltools.converters.mil import mil
from coremltools.converters.mil.mil import Builder as mb
from coremltools.converters.mil.mil import Function, Program, types
from coremltools.converters.mil._deployment_compatibility import AvailableTarget
from coremltools.converters.mil.mil.ops.defs._utils import (
    promote_input_dtypes,
)
from .utils import index_by_slices, update_tensor_by_slice, iterate_indexes_in_shapes, inverse_permutation
from .passes.utils import register_optimizations

from jaxlib.mlir import ir
from jaxlib.mlir.dialects.func import FuncOp, CallOp, ReturnOp as FuncReturnOp
from jaxlib.mlir.dialects.stablehlo import (
    AddOp, SubtractOp, MulOp, DivOp, NegOp, SignOp, AbsOp, ExpOp, Expm1Op, LogOp,
    Log1pOp, SqrtOp, ConstantOp, DotGeneralOp, ReshapeOp, BroadcastInDimOp, WhileOp,
    CompareOp, ConvertOp, SelectOp, DynamicSliceOp, ReturnOp, ConvolutionOp, MinOp,
    MaxOp, RsqrtOp, TanhOp, SineOp, CosineOp, TanOp, Atan2Op, ConcatenateOp, TransposeOp,
    DynamicUpdateSliceOp, SliceOp, CustomCallOp, IotaOp, ReduceOp, OrOp, AndOp, ReverseOp
)
from jaxlib.mlir.dialects.mhlo import (TopKOp)
from jax._src.lib.mlir.dialects import hlo

import numpy as np

from typing import List, Optional
import inspect
from functools import partial, reduce


def convert(module, minimum_deployment_target: AvailableTarget):
    if minimum_deployment_target < AvailableTarget.iOS18:
        raise ValueError("Converting to <iOS18 is not supported")

    register_optimizations()

    converter = StableHloConverter(opset_version=minimum_deployment_target)
    return converter.convert(module)


class TranscriptionContext:
    def __init__(self):
        self._path = []
        self.seen_paths = set()
        self.variables = {}  # Nested map: path -> variable -> mil var

    def push_function(self, name: str):
        counter = 0
        ctx_name = name
        while True:
            new_path = self._path + [ctx_name]
            if "/".join(new_path) in self.seen_paths:
                # Ensure that the new context name is in fact unique
                # A collision can happen if the same function is called twice
                ctx_name = f"{name}_{counter}"
                counter += 1
            else:
                self._path.append(ctx_name)
                self.seen_paths.add(self.path())
                return ctx_name

    def pop_function(self):
        self.variables.pop(self.path())
        self._path.pop()

    def add_variable(self, name: str, mil_var):
        path = self.path()
        if path not in self.variables:
            self.variables[path] = {}

        if name in self.variables[path]:
            raise ValueError(f"Variable {name} is already defined in path {path}")
        self.variables[path][name] = mil_var

    def __getitem__(self, name: str):
        # Walk up along the path list to find the first correctly named variable in scope
        path = self._path.copy()
        while True:
            ctx = self.variables["/".join(path)]
            if name in ctx:
                return ctx[name]
            if len(path) == 0:
                raise ValueError(f"Variable with name {name} is not defined in path {path}")
            path.pop()

    def path(self) -> str:
        return "/".join(self._path)


def register_stablehlo_op(func):
    # Check the signature
    sig = inspect.signature(func)
    params = list(sig.parameters.values())

    # Exclude 'self' from the parameters
    params = params[1:]

    error_msg = "HLO op implementations should take parameters of exactly " \
                "(context: TranscriptionContext, op: <HLO_OP_TYPE>)"
    if len(params) != 2:
        raise TypeError(error_msg)

    if not issubclass(params[0].annotation, TranscriptionContext):
        raise TypeError(error_msg)

    # We identify the function by the type of operation it implements
    func._implements_hlo_op = params[1].annotation
    return func


class StableHloOpsRegistry(type):
    def __init__(cls, name, bases, clsdict):
        super().__init__(name, bases, clsdict)

        cls._stablehlo_ops_registry = {}
        for name, method in clsdict.items():
            op_type = getattr(method, '_implements_hlo_op', False)
            if callable(method) and op_type:
                if op_type in cls._stablehlo_ops_registry:
                    raise TypeError(f"StableHLO op {op_type} has been registered more than once!")
                cls._stablehlo_ops_registry[op_type] = method

    def _dispatch_op(cls, self, context: TranscriptionContext, op):
        if type(op) not in self._stablehlo_ops_registry:
            raise TypeError(f"The StableHLO op {type(op)} has not been implemented!")

        op_method = self._stablehlo_ops_registry[type(op)]
        return op_method(self, context, op)

    def __call__(cls, *args, **kwargs):
        # Register the dispatch_op method
        instance = super().__call__(*args, **kwargs)
        setattr(instance, 'dispatch_op', cls._dispatch_op)
        return instance


class StableHloConverter(metaclass=StableHloOpsRegistry):

    def __init__(self, opset_version: Optional[int] = None):
        self.opset_version = AvailableTarget(opset_version) if opset_version is not None else None
        self.prog = mil.Program()
        self.func_index = {}

    def convert(self, module: ir.Module) -> Program:
        logger.info("Converting graph.")

        # Build function index to resolve/inline HLO function calls
        for func in module.body:
            self.func_index[func.name.value] = func

        for func in module.body:
            if "public" == func.visibility.value:
                self.build_func(func)

        return self.prog

    def build_func(self, hlo_func: FuncOp):
        context = TranscriptionContext()  # Map from results to created variables

        func_inputs = {}
        for arg in hlo_func.arguments:
            shape = arg.type.shape
            if shape == []:
                shape = [1]

            func_inputs[arg.get_name()] = mb.placeholder(
                shape=shape, dtype=self.__get_dtype(arg.type.element_type)
            )

        with Function(func_inputs, opset_version=self.opset_version) as ssa_func:
            for name in func_inputs.keys():
                context.add_variable(name, ssa_func.inputs[name])

            ssa_func.set_outputs(self.process_block(context, hlo_func.body.blocks[0]))
            self.prog.add_function(hlo_func.name.value, ssa_func)

    def process_block(self, context: TranscriptionContext, block: ir.Block):
        outputs = None
        for op in block:
            # Convention: Only the "return" op is returning from its building function
            # TODO: Check that "return" is always the last node!
            ret = self.dispatch_op(self, context, op)
            if ret is not None:
                if outputs is not None:
                    raise ValueError("More than 1 return op in block!")
                outputs = ret
        return outputs

    @register_stablehlo_op
    def op_call(self, context: TranscriptionContext, op: CallOp):
        # We can not do function calls in MIL, so we have to inline the function

        # Get the argument mapping prior to entering the function context
        context_args = []

        for arg in op.operands:
            context_args.append(context[arg.get_name()])

        func_name = op.callee.value
        hlo_func = self.func_index[op.callee.value]
        params = hlo_func.arguments
        outputs = self.__invoke_hlo_function(context, func_name, params, hlo_func.body, context_args)

        # Configure return value
        for result, output in zip(op.results, outputs):
            context.add_variable(result.get_name(), output)

    @register_stablehlo_op
    def op_return(self, context: TranscriptionContext, op: ReturnOp):
        return [context[result.get_name()] for result in op.operands]

    @register_stablehlo_op
    def op_func_return(self, context: TranscriptionContext, op: FuncReturnOp):
        # The HLO / MLIR types for function return ops seem to be both in use
        # The behaviour and fields of the two types should be similar, so we
        # simply delegate to the HLO version
        return self.op_return(context, op)

    @register_stablehlo_op
    def op_add(self, context: TranscriptionContext, op: AddOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_op = mb.add(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_or(self, context: TranscriptionContext, op: OrOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_op = mb.logical_or(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_and(self, context: TranscriptionContext, op: AndOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_op = mb.logical_and(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_subtract(self, context: TranscriptionContext, op: SubtractOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_op = mb.sub(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_mul(self, context: TranscriptionContext, op: MulOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_op = mb.mul(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_div(self, context: TranscriptionContext, op: DivOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]

        # From HLO constraints we know the base-types should line up
        lhs_type = self.__resolve_type(lhs)
        rhs_type = self.__resolve_type(rhs)
        if lhs_type != rhs_type:
            raise ValueError(f"Division not supported for different types. lhs type: {lhs_type}, rhs type: {rhs_type}")
        if types.is_complex(lhs_type):
            raise ValueError("Complex numbers are not supported in MIL")

        if types.is_float(lhs_type):
            cml_op = mb.real_div(x=lhs, y=rhs)
        elif types.is_int(lhs_type):
            cml_op = mb.floor_div(x=lhs, y=rhs)
        else:
            raise ValueError(f"Unknown dtype {lhs_type}")

        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_neg(self, context: TranscriptionContext, op: NegOp):
        # TODO(knielsen): Consider unsigned and more exotic types
        operand = context[op.operand.get_name()]
        minus_one = np.array([-1], dtype=types.nptype_from_builtin(operand.dtype))
        cml_op = mb.mul(x=minus_one, y=operand)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_sign(self, context: TranscriptionContext, op: SignOp):
        operand = context[op.operand.get_name()]
        cml_op = mb.sign(x=operand)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_abs(self, context: TranscriptionContext, op: AbsOp):
        operand = context[op.operand.get_name()]
        cml_op = mb.abs(x=operand)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_log(self, context: TranscriptionContext, op: LogOp):
        operand = context[op.operand.get_name()]
        cml_op = mb.log(x=operand)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_log1p(self, context: TranscriptionContext, op: Log1pOp):
        operand = context[op.operand.get_name()]
        one = np.array([1], dtype=types.nptype_from_builtin(self.__resolve_type(operand)))
        x_plus_one = mb.add(x=one, y=operand)
        cml_op = mb.log(x=x_plus_one)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_exp(self, context: TranscriptionContext, op: ExpOp):
        operand = context[op.operand.get_name()]
        cml_op = mb.exp(x=operand)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_expm1(self, context: TranscriptionContext, op: Expm1Op):
        operand = context[op.operand.get_name()]
        cml_op = mb.add(x=mb.exp(x=operand), y=-1.0)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_transpose(self, context: TranscriptionContext, op: TransposeOp):
        operand = context[op.operand.get_name()]
        perm = np.array(op.permutation, dtype=np.int32)
        cml_op = mb.transpose(x=operand, perm=perm)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_sqrt(self, context: TranscriptionContext, op: SqrtOp):
        operand = context[op.operand.get_name()]
        cml_op = mb.sqrt(x=operand)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_constant(self, context: TranscriptionContext, op: ConstantOp):
        constant = np.array(op.value)
        constant = np.reshape(constant, op.result.type.shape)
        context.add_variable(op.result.get_name(), constant)

    @register_stablehlo_op
    def op_dot_general(self, context: TranscriptionContext, op: DotGeneralOp):
        # This roughly follows the steps from https://github.com/openxla/stablehlo/blob/main/docs/spec.md#dot_general
        # but uses that we have a matrix multiplication primitive, instead of just a dot-product primitive.
        lhs_rank = len(op.lhs.type.shape)
        rhs_rank = len(op.rhs.type.shape)
        dot_dim_numbers = hlo.DotDimensionNumbers(op.dot_dimension_numbers)

        lhs_contracting_dim = dot_dim_numbers.lhs_contracting_dimensions
        rhs_contracting_dim = dot_dim_numbers.rhs_contracting_dimensions
        lhs_batching_dim = dot_dim_numbers.lhs_batching_dimensions
        rhs_batching_dim = dot_dim_numbers.rhs_batching_dimensions

        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]

        def multiply(lst: List):
            return reduce(lambda a, b: int(a) * int(b), lst, 1)

        def last_column_dot(lhs, rhs):
            # TODO: Figure out if we need to special case broadcasting dims
            return mb.matmul(x=lhs, y=rhs, transpose_y=True)

        # Remark: There is a potential performance optimization here:
        #         If we move the largest result dimensions of the tensor towards
        #         the end of the array, we may save a lot of work when iterating
        #         over the result indexes later, as the last dims will be handled
        #         by matrix multiplication
        lhs_result_dim = [dim for dim in range(lhs_rank) if dim not in lhs_batching_dim + lhs_contracting_dim]
        rhs_result_dim = [dim for dim in range(rhs_rank) if dim not in rhs_batching_dim + rhs_contracting_dim]

        # For both the lhs and rhs, put the dimensions being contracted last
        transposed_lhs = mb.transpose(x=lhs, perm=lhs_batching_dim + lhs_result_dim + lhs_contracting_dim)
        transposed_rhs = mb.transpose(x=rhs, perm=rhs_batching_dim + rhs_result_dim + rhs_contracting_dim)

        # Calculate the result by looping over the contracting dims in order
        result_shape = [lhs.shape[dim] for dim in lhs_batching_dim]
        result_shape += [lhs.shape[dim] for dim in lhs_result_dim]
        result_shape += [rhs.shape[dim] for dim in rhs_result_dim]
        if len(result_shape) == 0:
            # Special case for scalar result
            result_shape = [1]
        result = mb.fill(shape=result_shape)

        def calculate_result_index(lhs_idx, rhs_idx, acc):
            contracted_element_count = multiply([lhs.shape[dim] for dim in lhs_contracting_dim])
            # print(f"contracted_element_count = {contracted_element_count}")
            batch_selector = tuple([slice(None) for _i in range(len(lhs_batching_dim))])
            batch_shape = tuple([lhs.shape[dim] for dim in lhs_batching_dim])

            # Reshape the lhs and rhs to have all the contracting dimensions in the end.
            # We will always make them have the shape `(batch_shape, last_dim_shape, contraction_count)``
            # where we may have to set `last_dim_shape` to 1, if the dimension does not exist.
            lhs_for_result_idx = index_by_slices(transposed_lhs, list(batch_selector) + [lhs_idx, ...])
            if len(lhs_result_dim) > 0:
                lhs_reshape_shape = batch_shape + (lhs.shape[lhs_result_dim[-1]],) + (contracted_element_count, )
            else:
                lhs_reshape_shape = batch_shape + (1, contracted_element_count)
            contracted_lhs = mb.reshape(x=lhs_for_result_idx, shape=lhs_reshape_shape)

            rhs_for_result_idx = index_by_slices(transposed_rhs, list(batch_selector) + [rhs_idx, ...])
            if len(rhs_result_dim) > 0:
                rhs_reshape_shape = batch_shape + (rhs.shape[rhs_result_dim[-1]],) + (contracted_element_count, )
            else:
                rhs_reshape_shape = batch_shape + (1, contracted_element_count)
            contracted_rhs = mb.reshape(x=rhs_for_result_idx, shape=rhs_reshape_shape)

            # print(f"contracted_lhs shape: {contracted_lhs.shape}")
            # print(f"contracted_rhs shape: {contracted_rhs.shape}")

            idx_result = last_column_dot(contracted_lhs, contracted_rhs)

            # If we added a fake dimension, we will make sure to squeeze it away
            if len(lhs_result_dim) == 0 and len(rhs_result_dim) == 0:
                idx_result = mb.squeeze(x=idx_result, axes=(-1, -2))
            elif len(lhs_result_dim) == 0:
                idx_result = mb.squeeze(x=idx_result, axes=(-2,))
            elif len(rhs_result_dim) == 0:
                idx_result = mb.squeeze(x=idx_result, axes=(-1,))

            # TODO: Consider making this work on iOS<18 by using concatenation
            # We may have to add an extra slice for the skipped dimension
            result_idx = []
            result_idx.append(lhs_idx)
            if len(lhs_result_dim) > 0:
                result_idx.append(slice(None))
            result_idx.append(rhs_idx)
            if len(rhs_result_dim) > 0:
                result_idx.append(slice(None))

            return [update_tensor_by_slice(acc, list(batch_selector) + result_idx, idx_result)]

        # We can utilize that we have a full matrix multiply primitive available, compared to having only
        # a dot-product primitive. Therefore we can avoid iterating over the last dimension in respectively
        # the lhs and rhs tensors
        lhs_shape = [lhs.shape[dim] for dim in lhs_result_dim[:-1]]
        rhs_shape = [rhs.shape[dim] for dim in rhs_result_dim[:-1]]
        # In principle all of the matrix multiplications generated here, could be done in parallel.
        # MIL does not seem to support this.
        # We could try to combine the matrix multiplications when the shapes allow it, but for now
        # we will just loop through them sequentially.
        result, = iterate_indexes_in_shapes(calculate_result_index, [lhs_shape, rhs_shape], [result])

        context.add_variable(op.result.get_name(), result)

    @register_stablehlo_op
    def op_reshape(self, context: TranscriptionContext, op: ReshapeOp):
        x = context[op.operand.get_name()]
        new_shape = op.result.type.shape
        reshape_res = mb.reshape(x=x, shape=new_shape)
        context.add_variable(op.result.get_name(), reshape_res)

    @register_stablehlo_op
    def op_broadcast_in_dim(self, context: TranscriptionContext, op: BroadcastInDimOp):
        x = context[op.operand.get_name()]

        reshaped_operand_shape = [1] * len(op.result.type.shape)
        for i, op_shape in enumerate(op.operand.type.shape):
            result_idx = op.broadcast_dimensions[i]
            reshaped_operand_shape[result_idx] = op_shape

        x = mb.reshape(x=x, shape=reshaped_operand_shape)
        for result_dim, current_shape in enumerate(reshaped_operand_shape):
            if current_shape != op.result.type.shape[result_dim]:
                assert current_shape == 1
                # Replicate data along dimension `dim` until the result dimension is filled up
                values = [x] * op.result.type.shape[result_dim]
                x = mb.concat(values=values, axis=result_dim)

        context.add_variable(op.result.get_name(), x)

    @register_stablehlo_op
    def op_while(self, context: TranscriptionContext, op: WhileOp):
        def cond(*loop_args):
            params = [param for param in op.cond.blocks[0].arguments]
            outputs = self.__invoke_hlo_function(context, "while_cond", params, op.cond, loop_args)
            if len(outputs) != 1:
                raise ValueError("The output of while_cond should always be a single boolean!")
            # TODO(knielsen): Add a check that the output is in fact a single boolean value

            return outputs[0]

        def body(*body_args):
            params = [param for param in op.body.blocks[0].arguments]
            return self.__invoke_hlo_function(context, "while_body", params, op.body, body_args)

        loop_vars = [context[arg.get_name()] for arg in op.operands]
        while_results = mb.while_loop(_cond=cond, _body=body, loop_vars=loop_vars)

        for result_var, while_result in zip(op.results, while_results):
            context.add_variable(result_var.get_name(), while_result)

    @register_stablehlo_op
    def op_compare(self, context: TranscriptionContext, op: CompareOp):
        comparison_direction = hlo.ComparisonDirectionAttr(op.comparison_direction).value
        cml_op_builder = {
            "EQ": mb.equal,
            "NE": mb.not_equal,
            "GE": mb.greater_equal,
            "GT": mb.greater,
            "LE": mb.less_equal,
            "LT": mb.less,
        }[comparison_direction]

        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_op = cml_op_builder(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_convert(self, context: TranscriptionContext, op: ConvertOp):
        x = context[op.operand.get_name()]
        new_dtype = self.__get_dtype(op.result.type.element_type)
        cml_op = mb.cast(x=x, dtype=self.__dtype_str(new_dtype))
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_select(self, context: TranscriptionContext, op: SelectOp):
        cond = context[op.pred.get_name()]
        a = context[op.on_true.get_name()]
        b = context[op.on_false.get_name()]
        cml_op = mb.select(cond=cond, a=a, b=b)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_dynamic_slice(self, context: TranscriptionContext, op: DynamicSliceOp):
        x = context[op.operand.get_name()]

        # The HLO DynamicSliceOp gives the start indices as seperate 0-dimensional integer variables
        # We need to convert them to a tensor to be compatible with mb.slice_by_size
        start_idx_variables = [context[i.get_name()] for i in op.start_indices]
        begin = mb.concat(values=start_idx_variables, axis=0)

        # The slice sizes in HLO are given by a signed integer with 64 bits
        # This is not supported by MIL, so we convert it to a MIL int32 type
        # TODO(knielsen): Overflow check?
        sizes = np.array(op.slice_sizes, dtype=np.int32)

        cml_op = mb.slice_by_size(x=x, begin=begin, size=sizes)
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_slice(self, context: TranscriptionContext, op: SliceOp):
        x = context[op.operand.get_name()]

        begin = np.array(op.start_indices, dtype=np.int32)
        end = np.array(op.limit_indices, dtype=np.int32)
        stride = np.array(op.strides, dtype=np.int32)

        cml_op = mb.slice_by_index(
            x=x,
            begin=begin,
            end=end,
            stride=stride,
        )
        context.add_variable(op.result.get_name(), cml_op)

    @register_stablehlo_op
    def op_dynamic_update_slice(self, context: TranscriptionContext, op: DynamicUpdateSliceOp):
        x = context[op.operand.get_name()]
        updates = context[op.update.get_name()]

        start_indices = [context[i.get_name()] for i in op.start_indices]
        start_indices = mb.concat(values=start_indices, axis=0)
        end_indices = mb.add(x=start_indices, y=op.update.type.shape)

        update_res = mb.slice_update(
            x=x,
            update=updates,
            begin=start_indices,
            end=end_indices,
        )
        context.add_variable(op.result.get_name(), update_res)

    @register_stablehlo_op
    def op_convolution(self, context: TranscriptionContext, op: ConvolutionOp):
        dim_spec = hlo.ConvDimensionNumbers(op.dimension_numbers)
        # TODO(knielsen): It should be possible to remove this batch dimension check, but
        #                 there should be a unit test testing it.
        if dim_spec.input_batch_dimension != 0 or dim_spec.output_batch_dimension != 0:
            raise ValueError(f"Only the first dimension is currently supported for batch dimension. Got {dim_spec}")
        if len(dim_spec.input_spatial_dimensions) > 3 or len(dim_spec.output_spatial_dimensions) > 3:
            raise ValueError("MIL only supports convolutions with dim <= 3")

        if op.batch_group_count.value != 1:
            raise ValueError(f"Only a batch group count of 1 is supported. Got {op.batch_group_count.value}")

        # MIL expects it on the form [input_batch_dimension, input_feature_dimension, spatial_dimensions*]
        input_permutation = [
            dim_spec.input_batch_dimension,
            dim_spec.input_feature_dimension,
            *dim_spec.input_spatial_dimensions
        ]
        x = context[op.lhs.get_name()]  # The inputs comes from vars
        x = mb.transpose(x=x, perm=input_permutation)

        strides = None
        if op.window_strides is not None:
            strides = np.array(op.window_strides, dtype=np.int32)

        kernel_dilation = None
        if op.rhs_dilation is not None:
            kernel_dilation = np.array(op.rhs_dilation, dtype=np.int32)

        groups = op.feature_group_count.value

        # Handle padding
        # TODO(knielsen): Consider moving splat/non-splat handling to some utility
        in_rank = x.rank - 2
        if op.padding is None:
            pad = np.zeros((2 * in_rank), dtype=np.int32)
        elif op.padding.is_splat:
            pad = op.padding.get_splat_value().value * np.ones((2 * in_rank), dtype=np.int32)
        else:
            # We need to reshape the array to a linear array to match MILs expectation
            pad = np.reshape(np.array(op.padding, dtype=np.int32), (2 * in_rank, ))

        # We switch the convolution to a transposed convolution if we have lhs_dilation
        conv_type = mb.conv
        if op.lhs_dilation:
            lhs_dilations = np.array(op.lhs_dilation, dtype=np.int32)
            if np.any(lhs_dilations > 1):
                # This is a transpoed convolution
                if strides is not None:
                    raise ValueError("For a conv with lhs dilation we expect the stride to be not set! "
                                     "Because convolution with input dilation d is equivalent to transposed "
                                     "convolution with stride d.")
                # Convolution with input dilation d is equivalent to transposed convolution with stride d
                strides = lhs_dilations

                output_shape = op.result.type.shape
                output_shape.append(output_shape.pop(1))  # Match the format of MIL

                conv_type = partial(
                    mb.conv_transpose,
                    output_shape=output_shape
                )

                # We need to subtract 1 from the padding to make the dimensions line up
                pad -= 1
                if np.any(pad < 0):
                    raise ValueError("The case where the padding turns negative when translating to a "
                                     "transposed convolution is not supported.")

        # The MIL weights should be on form:
        #  - normal convolutions: [output_features, input_features / groups, spatial kernels*]
        #  - transposed convolutions: [input_features, output_features / groups, spatial kernels*]
        weight = context[op.rhs.get_name()]  # The weights are numpy arrays
        weight_permutation = []
        if conv_type == mb.conv:
            weight_permutation = [
                dim_spec.kernel_output_feature_dimension,
                dim_spec.kernel_input_feature_dimension,
                *dim_spec.kernel_spatial_dimensions
            ]
        else:
            weight_permutation = [
                dim_spec.kernel_input_feature_dimension,
                dim_spec.kernel_output_feature_dimension,
                *dim_spec.kernel_spatial_dimensions
            ]
        weight = mb.transpose(x=weight, perm=weight_permutation)

        # TODO(knielsen): Make this check more readable!
        # It is executed for conv transpose
        if conv_type != mb.conv:
            # MIL expects the weights to be reversed along the kernel dimensions
            kernel_dimensions = [i + 2 for i in range(len(weight.shape) - 2)]
            weight = mb.reverse(x=weight, axes=kernel_dimensions)

        cml_conv = conv_type(
            x=x,
            weight=weight,
            strides=strides,
            pad_type="custom",
            pad=pad,
            dilations=kernel_dilation,
            groups=groups,
        )

        # Re-arrange output dimensions to match expectation
        # MIL outputs on the form [batch, features, spatial dims*]
        output_permutation = inverse_permutation([
            dim_spec.output_batch_dimension,
            dim_spec.output_feature_dimension,
            *dim_spec.output_spatial_dimensions
        ])
        cml_conv = mb.transpose(x=cml_conv, perm=output_permutation)

        context.add_variable(op.result.get_name(), cml_conv)

    @register_stablehlo_op
    def op_max(self, context: TranscriptionContext, op: MaxOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_res = mb.maximum(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_res)

    @register_stablehlo_op
    def op_min(self, context: TranscriptionContext, op: MinOp):
        lhs = context[op.lhs.get_name()]
        rhs = context[op.rhs.get_name()]
        cml_res = mb.minimum(x=lhs, y=rhs)
        context.add_variable(op.result.get_name(), cml_res)

    @register_stablehlo_op
    def op_rsqrt(self, context: TranscriptionContext, op: RsqrtOp):
        x = context[op.operand.get_name()]
        mil_res = mb.rsqrt(x=x)
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_tanh(self, context: TranscriptionContext, op: TanhOp):
        x = context[op.operand.get_name()]
        mil_res = mb.tanh(x=x)
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_sine(self, context: TranscriptionContext, op: SineOp):
        x = context[op.operand.get_name()]
        mil_res = mb.sin(x=x)
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_cosine(self, context: TranscriptionContext, op: CosineOp):
        x = context[op.operand.get_name()]
        mil_res = mb.cos(x=x)
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_tan(self, context: TranscriptionContext, op: TanOp):
        x = context[op.operand.get_name()]
        mil_res = mb.tan(x=x)
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_atan2(self, context: TranscriptionContext, op: Atan2Op):
        y = context[op.lhs.get_name()]
        x = context[op.rhs.get_name()]
        # Notice the fraction may be +-inf
        fraction = mb.real_div(x=y, y=x)
        atan2_res = mb.atan(x=fraction)
        # We need to adjust for negative x, based on the sign of y
        atan2_res_adjusted = mb.add(x=atan2_res, y=mb.mul(x=mb.sign(x=y), y=np.pi))
        atan2_res = mb.select(
            cond=mb.less(x=x, y=0.0),
            a=atan2_res_adjusted,
            b=atan2_res,
        )
        context.add_variable(op.result.get_name(), atan2_res)

    @register_stablehlo_op
    def op_concatenate(self, context: TranscriptionContext, op: ConcatenateOp):
        values = [context[input.get_name()] for input in op.inputs]
        values = promote_input_dtypes(values)
        mil_res = mb.concat(values=values, axis=op.dimension.value)
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_reverse(self, context: TranscriptionContext, op: ReverseOp):
        x = context[op.operand.get_name()]
        mil_res = mb.reverse(x=x, axes=np.array(op.dimensions, dtype=np.int32))
        context.add_variable(op.result.get_name(), mil_res)

    @register_stablehlo_op
    def op_reduce(self, context: TranscriptionContext, op: ReduceOp):
        # HLO reductions can be arbitrarily complex and defines a custom function
        # specifying the reduction.
        # Unforunately this level of granularity is not supported through MIL.
        # We try to detect some simple cases for reductions mapping to native MIL
        # instructions, and otherwise fall back to a MIL while-loop based implementation.

        def match_reduction_type(hlo_body):
            if len(hlo_body.blocks) != 1:
                return None
            args = list(hlo_body.blocks[0].arguments)
            ops = list(hlo_body.blocks[0].operations)

            # Simple matches are where the `hlo_body` is on the form
            #   return _generic_reduction_op_type_(`args`)
            # In that case, if MIL has an equvalent of `_generic_reduction_op_`, we simply delegate to that
            simple_matches = {
                MaxOp: mb.reduce_max,
                MinOp: mb.reduce_min,
                AddOp: mb.reduce_sum,
                MulOp: mb.reduce_prod,
            }

            for generic_reduce_op_type, mil_equivalent in simple_matches.items():
                if len(ops) == 2 and isinstance(ops[0], generic_reduce_op_type) and isinstance(ops[1], ReturnOp):
                    if list(ops[0].operands) == args and list(ops[1].operands) == list(ops[0].results):
                        return mil_equivalent

            return None

        reduction_type = match_reduction_type(op.body)
        if reduction_type and len(op.inputs) == 1:
            input = context[op.inputs[0].get_name()]
            res = reduction_type(x=input, axes=np.array(op.dimensions, dtype=np.int32))
            context.add_variable(op.result.get_name(), res)
        else:
            logger.warn("Falling back to while-loop implementation for reduction. This may be slower than expected!")

            input_rank = len(op.inputs[0].type.shape)
            inputs = [context[input.get_name()] for input in op.inputs]
            # Notice for the loops we treat both `reduce_shape` and `result_shape` as being
            # of the input rank. This is to make computing element indexes easier.
            # When updating the result, we later pick out just the result indices
            # we care about in the actual result.
            reduce_shape = [inputs[0].shape[dim] if dim in op.dimensions else 1 for dim in range(input_rank)]
            result_shape = [inputs[0].shape[dim] if dim not in op.dimensions else 1 for dim in range(input_rank)]
            init_values = [context[init_value.get_name()] for init_value in op.init_values]

            def compute_reduction(result_idx, *partial_results):
                def compute_inner(element_idx, *acc):
                    element_idx = mb.add(x=result_idx, y=element_idx)
                    elements = [mb.reshape(x=index_by_slices(input, [element_idx]), shape=(1,)) for input in inputs]

                    args = list(acc) + elements
                    hlo_params = list(op.body.blocks[0].arguments)
                    outputs = self.__invoke_hlo_function(context, "reduce_body", hlo_params, op.body, args)

                    return outputs

                reduction_results = iterate_indexes_in_shapes(compute_inner, [reduce_shape], init_values)

                # The result rank is likely less than the input shape.
                # We need to pick the indexes in the result shape we want to update
                result_indices = [dim for dim in range(input_rank) if dim not in op.dimensions]
                if len(result_indices) != 0:
                    result_idx = [mb.gather(x=result_idx, indices=result_indices)]
                else:
                    result_idx = []

                return [
                    update_tensor_by_slice(acc, result_idx, result)
                    for acc, result in zip(partial_results, reduction_results)
                ]

            mil_results = [
                np.zeros(result.type.shape, dtype=types.nptype_from_builtin(self.__get_dtype(result.type.element_type)))
                for result in op.results
            ]
            mil_results = iterate_indexes_in_shapes(compute_reduction, [result_shape], mil_results, unroll_limit=5)
            for (res, mil_res) in zip(op.results, mil_results):
                context.add_variable(res.get_name(), mil_res)

    @register_stablehlo_op
    def op_iota(self, context: TranscriptionContext, op: IotaOp):
        iota_dim = int(op.iota_dimension)
        tensor_shape = op.result.type.shape
        vec_shape = [tensor_shape[dim] if dim == iota_dim else 1 for dim in range(len(tensor_shape))]
        dtype = types.nptype_from_builtin(self.__get_dtype(op.result.type.element_type))
        res = np.reshape(np.arange(tensor_shape[iota_dim], dtype=dtype), vec_shape) * np.ones(tensor_shape, dtype=dtype)
        context.add_variable(op.result.get_name(), res)

    @register_stablehlo_op
    def op_custom_call(self, context: TranscriptionContext, op: CustomCallOp):
        if op.call_target_name.value.startswith("mhlo."):
            mapped_op = None
            op_impl = None
            match op.call_target_name.value:
                case "mhlo.topk":
                    mapped_op = TopKOp
                    op_impl = self._op_mhlo_topk

            if not mapped_op:
                raise ValueError(f"mhlo op '{op.call_target_name.value}' is not implemented")
            if not op_impl:
                raise ValueError(f"mhlo op '{op.call_target_name.value}' does not have an implementation")

            mhlo_attributes = {attr.name: attr.attr for attr in list(op.attributes["mhlo.attributes"])}
            delegate_op = partial(mapped_op, **mhlo_attributes, loc=op.location)(*op.operands)

            # We manually have to handle the results, as the current API does not allow naming
            # the `delegate_op` results according to the custom call results
            mil_results = op_impl(context, delegate_op)
            for (custom_call_result, mil_result) in zip(op.results, mil_results):
                context.add_variable(custom_call_result.get_name(), mil_result)

            return

        raise ValueError(f"Custom call is not supported: {op.call_target_name}")

    def _op_mhlo_topk(self, context: TranscriptionContext, op: TopKOp):
        """
        This is a MHLO op, and follows a slightly different pattern, since it is unvoked by a
        custom call. It will return the results, as we currently can not rename the results
        in the TopKOp
        """
        x = context[op.operand.get_name()]
        mil_res = mb.topk(x=x, k=op.k.value, ascending=not op.largest.value)
        return mil_res

    def __invoke_hlo_function(self, context: TranscriptionContext, func_name: str, hlo_params, hlo_func_body, cml_args):
        # Enter variable context for the function call
        context.push_function(func_name)

        # Setup arguments for the function
        for hlo_func_param, actual_arg in zip(hlo_params, cml_args):
            context.add_variable(hlo_func_param.get_name(), actual_arg)

        # Process the function
        if len(hlo_func_body.blocks) != 1:
            raise ValueError(f"Unsupported function with {len(hlo_func_body.blocks)} blocks")
        outputs = self.process_block(context, hlo_func_body.blocks[0])

        # Exit the function context
        context.pop_function()

        return outputs

    def __resolve_type(self, obj):
        if isinstance(obj, np.ndarray):
            return types.numpy_type_to_builtin_type(obj.dtype)
        return obj.dtype

    def __dtype_str(self, type):
        # TODO(knielsen): Add additional types
        return {
            types.int32: "int32",
            types.fp16: "fp16",
            types.fp32: "fp32",
        }[type]

    def __get_dtype(self, element_type):
        if isinstance(element_type, ir.IntegerType):
            # TODO(knielsen): Handle different kinds of integer types
            return types.int32
        if isinstance(element_type, ir.F16Type):
            return types.fp16
        if isinstance(element_type, ir.F32Type):
            return types.fp32
        raise ValueError(f"Unsupported type {element_type}")



import jax
from jax.export import export as _jax_export
import jax.numpy as jnp
from jax._src.lib.mlir import ir
from jax._src.interpreters import mlir as jax_mlir
import numpy as np

from stablehlo_coreml.converter import convert
from stablehlo_coreml import DEFAULT_HLO_PIPELINE
import coremltools as ct
from coremltools.converters.mil.testing_utils import compare_backend
from coremltools.converters.mil.mil import Program, Block

from functools import partial
from typing import List


def test_addition():
    run_and_compare(jnp.add, (jnp.float32(1), jnp.float32(1)))
    run_and_compare(jnp.add, (jnp.zeros((2, 2, 2)), jnp.zeros((2, 2, 2))))


def test_tensor_multiplication():
    def scalar_product(lhs, rhs):
        return jnp.einsum("a,a", lhs, rhs)

    def scalar_with_vector(lhs, rhs):
        return jnp.einsum("a,b->ab", lhs, rhs)

    def scalar_with_matrix(lhs, rhs):
        return jnp.einsum("a,bc->abc", lhs, rhs)

    def vector_with_matrix(lhs, rhs):
        return jnp.einsum("a,ab->b", lhs, rhs)

    def matrix_multiplication(lhs, rhs):
        return jnp.einsum("ij,jk -> ik", lhs, rhs)

    def outer_product_with_single_batch_dim(lhs, rhs):
        return jnp.einsum("abc,ajk->abcjk", lhs, rhs)

    def single_contraction_single_batch(lhs, rhs):
        return jnp.einsum("abcd,ackl->abdkl", lhs, rhs)

    def two_contractions_single_batch(lhs, rhs):
        return jnp.einsum("abcd,ackd->abk", lhs, rhs)

    def three_contractions_single_batch(lhs, rhs):
        return jnp.einsum("abcd,acbd->a", lhs, rhs)

    def contract_all(lhs, rhs):
        return jnp.einsum("abcd,acbd", lhs, rhs)

    def full_tensor_product(lhs, rhs):
        return jnp.einsum("ab,ihj->abihj", lhs, rhs)

    def full_tensor_product_1_4(lhs, rhs):
        return jnp.einsum("a,ihjk->aihjk", lhs, rhs)

    def full_tensor_product_3_2(lhs, rhs):
        return jnp.einsum("abc,ih->abcih", lhs, rhs)

    def full_tensor_product_4_1(lhs, rhs):
        return jnp.einsum("abcd,i->abcdi", lhs, rhs)

    run_and_compare(scalar_product, (jnp.zeros((1)), jnp.zeros((1))))
    run_and_compare(scalar_with_vector, (jnp.zeros((1)), jnp.zeros((5))))
    run_and_compare(scalar_with_matrix, (jnp.zeros((1)), jnp.zeros((5, 3))))
    run_and_compare(vector_with_matrix, (jnp.zeros((5)), jnp.zeros((5, 3))))
    run_and_compare(matrix_multiplication, (jnp.zeros((3, 4)), jnp.zeros((4, 5))))
    run_and_compare(outer_product_with_single_batch_dim, (jnp.zeros((2, 3, 4)), jnp.zeros((2, 4, 5))))
    run_and_compare(single_contraction_single_batch, (jnp.zeros((2, 3, 4, 5)), jnp.zeros((2, 4, 2, 5))))
    run_and_compare(two_contractions_single_batch, (jnp.zeros((2, 3, 4, 5)), jnp.zeros((2, 4, 2, 5))))
    run_and_compare(three_contractions_single_batch, (jnp.zeros((2, 3, 4, 5)), jnp.zeros((2, 4, 3, 5))))
    run_and_compare(contract_all, (jnp.zeros((2, 3, 4, 5)), jnp.zeros((2, 4, 3, 5))))
    run_and_compare(full_tensor_product, (jnp.zeros((2, 3)), jnp.zeros((2, 4, 3))))

    # Test the full tensor product with a big dimensions, and ensure that the program gets handled by a dynamic loop
    run_and_compare(full_tensor_product, (jnp.zeros((10, 3)), jnp.zeros((15, 20, 3))))
    run_and_compare(full_tensor_product_1_4, (jnp.zeros((10,)), jnp.zeros((15, 20, 5, 3))))
    run_and_compare(full_tensor_product_1_4, (jnp.zeros((2,)), jnp.zeros((2, 2, 2, 3))))
    run_and_compare(full_tensor_product_3_2, (jnp.zeros((20, 10, 3)), jnp.zeros((15, 3))))
    run_and_compare(full_tensor_product_3_2, (jnp.zeros((2, 2, 3)), jnp.zeros((2, 3))))
    run_and_compare(full_tensor_product_4_1, (jnp.zeros(((15, 20, 5, 3))), jnp.zeros((10,))))
    run_and_compare(full_tensor_product_4_1, (jnp.zeros(((2, 2, 2, 3))), jnp.zeros((2,))))


def test_simple_reductions():
    def compare_and_ensure_no_loops(jax_func, input_spec):
        cml_model = run_and_compare(jax_func, input_spec)
        assert "while_loop" not in get_model_instruction_types(cml_model)

    compare_and_ensure_no_loops(partial(jnp.max, axis=1), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.max, axis=1, keepdims=True), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.sum, axis=0), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.sum, axis=1), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.sum, axis=2), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.sum, axis=(0, 2)), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.sum, axis=(0, 1, 2)), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.min, axis=0), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.min, axis=(1, 2)), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.mean, axis=0), (jnp.zeros((2, 3, 4)),))
    compare_and_ensure_no_loops(partial(jnp.prod, axis=1), (jnp.zeros((2, 3, 4)),))


def test_complex_reductions():
    """
    These reductions are complicated, and will be handled using while loops (potentially unrolled)
    """
    run_and_compare(jnp.argmax, (jnp.zeros((2, 3, 3)),))
    run_and_compare(partial(jnp.argmax, keepdims=True), (jnp.zeros((2, 3, 3)),))
    run_and_compare(jnp.argmax, (jnp.zeros((20, 30, 40)),))
    run_and_compare(partial(jnp.argmax, keepdims=True), (jnp.zeros((20, 30, 40)),))
    run_and_compare(partial(jnp.argmax, axis=1), (jnp.zeros((2, 3, 3)),))
    run_and_compare(partial(jnp.argmax, axis=0, keepdims=True), (jnp.zeros((2, 3, 3)),))
    run_and_compare(partial(jnp.argmax, axis=2), (jnp.zeros((2, 3, 3)),))
    run_and_compare(partial(jnp.argmax, axis=1), (jnp.zeros((20, 30, 40)),))

    run_and_compare(jnp.argmin, (jnp.zeros((2, 3, 3)),))
    run_and_compare(partial(jnp.argmin, axis=1), (jnp.zeros((2, 3, 3)),))
    run_and_compare(partial(jnp.argmin, axis=1), (jnp.zeros((20, 30, 40)),))
    run_and_compare(partial(jnp.argmin, axis=1, keepdims=True), (jnp.zeros((20, 30, 40)),))


def test_topk():
    input_shape = (3, 5, 10)
    run_and_compare(partial(jax.lax.top_k, k=3), (jnp.zeros(input_shape),))


def test_reverse():
    run_and_compare(jnp.flip, (jnp.zeros((5,)),))
    run_and_compare(jnp.flip, (jnp.zeros((5, 5, 5)),))
    run_and_compare(partial(jnp.flip, axis=(0, 2)), (jnp.zeros((5, 5, 5)),))
    run_and_compare(partial(jnp.flip, axis=(1,)), (jnp.zeros((5, 5, 5)),))


def test_trigonmetry():
    run_and_compare(jnp.sin, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.cos, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.tan, (jnp.zeros((5, 6)),))

    run_and_compare(jnp.arcsin, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.arccos, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.arctan, (jnp.zeros((5, 6)),))

    run_and_compare(jnp.sinh, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.cosh, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.tanh, (jnp.zeros((5, 6)),))

    run_and_compare(jnp.arcsinh, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.arccosh, (jnp.zeros((5, 6)),))
    run_and_compare(jnp.arctanh, (jnp.zeros((5, 6)),))

    run_and_compare(jnp.atan2, (jnp.zeros((50, 20)), jnp.zeros((50, 20)),))


def jax_export(jax_func, input_spec):
    def compute_input_shapes(input_specs):
        shapes = []
        for input_spec in input_specs:
            if isinstance(input_spec, (list, tuple)):
                # We only unwrap the shapes for one level
                shapes.append(input_spec)
            else:
                shapes.append(jax.ShapeDtypeStruct(input_spec.shape, input_spec.dtype))
        return shapes
    input_shapes = compute_input_shapes(input_spec)
    jax_exported = _jax_export(jax.jit(jax_func))(*input_shapes)
    return jax_exported


def generate_random_from_shape(input_spec, key=jax.random.PRNGKey):
    shape = input_spec.shape
    dtype = input_spec.dtype
    output = jax.random.uniform(key=key, shape=shape, dtype=dtype, minval=-10, maxval=10)
    return output


def flatten(nested_list):
    def visit(lst):
        flat = []
        for element in lst:
            if isinstance(element, (list, tuple)):
                flat += visit(element)
            else:
                flat.append(element)
        return flat
    return visit(nested_list)


def __nest_flat_jax_input_to_input_spec(input_spec, flat_input):
    idx = 0

    def visit(lst):
        nonlocal idx
        result = []
        for element in lst:
            if isinstance(element, (list, tuple)):
                result.append(visit(element))
            else:
                if idx >= len(flat_input):
                    raise ValueError(
                        "flat_input had too many inputs to fit input_spec. "
                        f"Input spec: {input_spec}, Flat input: {flat_input}")
                result.append(flat_input[idx])
                idx += 1
        return result

    structured_input = visit(input_spec)
    if idx != len(flat_input):
        raise ValueError("flat_input had too few inputs to fill input_spec. "
                         f"Input spec: {input_spec}, Flat input: {flat_input}")

    return structured_input


def _count_program_complexity(mil_program: Program):
    """
    Counts the number of instructions in the given `mil_program`
    This is used to ensure we don't generate crazy big programs
    """
    def count_block(block: Block):
        complexity = 0
        for op in block.operations:
            for child_block in op.blocks:
                complexity += count_block(child_block)
            complexity += 1
        return complexity

    total_complexity = 0
    for func in mil_program.functions.values():
        total_complexity += count_block(func)
    return total_complexity


def run_and_compare(jax_func, input_spec, max_complexity: int = 10_000):
    """
    Converts the given `jax_func` to a CoreML model.
    Both models will be run on random input data with shapes specified by `input_spec`.
    If the CoreML model and `jax_func` does not agree on the output, an error will be raised.
    The resulting CoreML model will be returned.
    """

    jax_func = jax.jit(jax_func)
    exported = jax_export(jax_func, input_spec)
    context = jax_mlir.make_ir_context()
    hlo_module = ir.Module.parse(exported.mlir_module(), context=context)
    # print(f"HLO module: {hlo_module}")

    mil_program = convert(hlo_module, minimum_deployment_target=ct.target.iOS18)
    program_complexity = _count_program_complexity(mil_program)
    if program_complexity > max_complexity:
        raise ValueError(
            f"Generated a MIL program with complexity {program_complexity}, "
            "max allowed complexity is {max_complexity}"
        )

    pipeline = DEFAULT_HLO_PIPELINE
    # We temporarily avoid fp16 conversions in tests because of https://github.com/apple/coremltools/issues/2324
    passes_to_remove = [
         'common::add_fp16_cast'
    ]
    pipeline.remove_passes(passes_to_remove)

    cml_model = ct.convert(
        mil_program,
        source="milinternal",
        minimum_deployment_target=ct.target.iOS18,
        pass_pipeline=pipeline,
    )

    # Generate random inputs that matches cml_model input spec
    cml_input_key_values = {}
    jax_input_values = []
    key = jax.random.PRNGKey(0)
    for input_name, input_shape in zip(cml_model.input_description, exported.in_avals):
        key, value_key = jax.random.split(key, num=2)
        input_value = generate_random_from_shape(input_shape, value_key)
        cml_input_key_values[input_name] = input_value
        jax_input_values.append(input_value)

    # Transfor the input to match the Jax model, and call it
    jax_input_values = __nest_flat_jax_input_to_input_spec(input_spec, jax_input_values)
    expected_output = jax_func(*jax_input_values)

    # TODO(knielsen): Is there a nicer way of doing this?
    if not isinstance(expected_output, (list, tuple)):
        expected_output = (expected_output, )

    # Prepare the output for comparison
    cml_expected_outputs = {}
    for output_name, output_value in zip(cml_model.output_description, flatten(expected_output)):
        cml_expected_outputs[output_name] = np.asarray(output_value)

    compare_backend(cml_model, cml_input_key_values, cml_expected_outputs)

    return cml_model


def get_model_instruction_types(cml_model) -> List[str]:
    def collect_ops(ops: List) -> List[str]:
        collected_ops = []
        for op in ops:
            collected_ops.append(op.op_type)
            for block in op.blocks:
                collected_ops += collect_ops(block.operations)

        return collected_ops

    mil_program = cml_model._mil_program
    all_ops = []
    for func in mil_program.functions.values():
        all_ops += collect_ops(func.operations)
    return all_ops



from flax import nnx
import jax
import jax.numpy as jnp
from jax._src.typing import Array
from jax._src import core
from jax._src import dtypes

from functools import partial
from typing import Any, Optional, List
import einops

KeyArray = Any
DTypeLikeFloat = Any
DTypeLikeComplex = Any
DTypeLikeInexact = Any  # DTypeLikeFloat | DTypeLikeComplex
RealNumeric = Any  # Scalar jnp array or float


def uniform(minval: RealNumeric = 0,
            maxval: RealNumeric = 1,
            dtype: DTypeLikeInexact = jnp.float_) -> nnx.Initializer:

    def init(key: KeyArray,
             shape: core.Shape,
             dtype: DTypeLikeInexact = dtype) -> Array:
        dtype = dtypes.canonicalize_dtype(dtype)
        return jax.random.uniform(key, shape, dtype, minval=minval, maxval=maxval)

    return init


class sLSTMCell(nnx.Module):
    cell_input_proj: nnx.Linear
    cell_state_proj: nnx.Linear

    input_proj: nnx.Linear
    input_state_proj: nnx.Linear

    forget_gate_proj: nnx.Linear
    forget_state_proj: nnx.Linear

    output_gate_proj: nnx.Linear
    output_state_proj: nnx.Linear

    if_conv: nnx.Conv

    def __init__(self, num_cells: int, rngs: nnx.Rngs, input_size: Optional[int] = None, apply_if_conv: bool = True):
        self.num_cells = num_cells
        construct_x_proj = partial(
            nnx.Linear,
            in_features=input_size if input_size else num_cells,
            out_features=num_cells,
            use_bias=True,
            rngs=rngs,
        )
        construct_hidden_proj = partial(
            nnx.Linear,
            in_features=num_cells,
            out_features=num_cells,
            use_bias=False,
            kernel_init=nnx.initializers.orthogonal(),
            rngs=rngs,
        )

        self.cell_input_proj = construct_x_proj()
        self.cell_state_proj = construct_hidden_proj()

        self.input_proj = construct_x_proj(
            bias_init=nnx.initializers.normal(1e-2)
        )
        self.input_state_proj = construct_hidden_proj()

        self.forget_gate_proj = construct_x_proj(
            bias_init=uniform(3.0, 6.0)
        )
        self.forget_state_proj = construct_hidden_proj()

        self.output_gate_proj = construct_x_proj()
        self.output_state_proj = construct_hidden_proj()

        self.if_conv = None
        if apply_if_conv:
            self.if_conv = nnx.Conv(
                in_features=1,
                out_features=1,
                kernel_size=4,
                rngs=rngs
            )

    def __call__(self, carry, x):
        cell_state, hidden_state, normalizer_state, stabilizer_state = carry

        # HACK: See init_carry for an explanation
        cell_state = einops.rearrange(cell_state, "(b h) -> b h", h=self.num_cells)
        hidden_state = einops.rearrange(hidden_state, "(b h) -> b h", h=self.num_cells)
        normalizer_state = einops.rearrange(normalizer_state, "(b h) -> b h", h=self.num_cells)
        stabilizer_state = einops.rearrange(stabilizer_state, "(b h) -> b h", h=self.num_cells)

        # print(f"Hidden state shape: {hidden_state.shape}")
        out = nnx.sigmoid(self.output_gate_proj(x) + self.output_state_proj(hidden_state))

        if_input = x
        if self.if_conv:
            if_input = self.if_conv(if_input[..., jnp.newaxis])
            if_input = jnp.squeeze(nnx.silu(if_input), axis=2)

        i_tilde = self.input_proj(if_input) + self.input_state_proj(hidden_state)
        # TODO: Consider trying a sigmoid forget activation as well!
        f_tilde = self.forget_gate_proj(if_input) + self.forget_state_proj(hidden_state)

        m = jnp.maximum(f_tilde + stabilizer_state, i_tilde)  # Stabilizer state
        i = jnp.exp(i_tilde - m)
        f = jnp.exp(f_tilde + stabilizer_state - m)

        z = nnx.tanh(self.cell_input_proj(x) + self.cell_state_proj(hidden_state))
        c = f * cell_state + i * z  # Cell state

        n = f * normalizer_state + i  # Normalizer state
        h = out * c / n  # Hidden state

        c = einops.rearrange(c, "b h -> (b h)")
        h_not_compacted = h
        h = einops.rearrange(h, "b h -> (b h)")
        n = einops.rearrange(n, "b h -> (b h)")
        m = einops.rearrange(m, "b h -> (b h)")

        return (c, h, n, m), h_not_compacted

    @classmethod
    def init_carry(cls, batch_size: int, num_cells: int, rngs: nnx.Rngs):
        initializer = partial(nnx.initializers.zeros, shape=(batch_size, num_cells), dtype=jnp.float16)
        c = initializer(key=rngs())
        # HACK: CoreML does not supprot tensors of rank > 5, so we squeeze the c carry to make it fit
        c = einops.rearrange(c, "b h -> (b h)")

        h = initializer(key=rngs())
        h = einops.rearrange(h, "b h -> (b h)")

        n = initializer(key=rngs())
        n = einops.rearrange(n, "b h -> (b h)")

        m = initializer(key=rngs())
        m = einops.rearrange(m, "b h -> (b h)")

        return c, h, n, m


class sLSTMBlock(nnx.Module):
    heads: List[sLSTMCell]
    input_norm: nnx.BatchNorm
    output_norm: nnx.BatchNorm

    up_proj_1: nnx.Linear
    up_proj_2: nnx.Linear
    down_proj: nnx.Linear

    def __init__(self, hidden_size: int, num_heads: int, rngs: nnx.Rngs):
        @nnx.split_rngs(splits=num_heads)
        @nnx.vmap(axis_size=num_heads)
        def create_heads(rngs: nnx.Rngs):
            return sLSTMCell(hidden_size, rngs)
        self.heads = create_heads(rngs)

        self.input_norm = nnx.BatchNorm(hidden_size, rngs=rngs)
        self.output_norm = nnx.BatchNorm(hidden_size, rngs=rngs)

        intermediate_size = int(hidden_size * num_heads * 4.0 / 3.0)
        self.up_proj_1 = nnx.Linear(
            in_features=hidden_size * num_heads,
            out_features=intermediate_size,
            rngs=rngs,
        )
        self.up_proj_2 = nnx.Linear(
            in_features=hidden_size * num_heads,
            out_features=intermediate_size,
            rngs=rngs,
        )
        self.down_proj = nnx.Linear(
            in_features=intermediate_size,
            out_features=hidden_size,
            rngs=rngs,
        )

    def __call__(self, carry, x):
        out = self.input_norm(x)

        head_def, head_states = nnx.split(self.heads)

        def eval_heads(head_state, carry, x):
            head = nnx.merge(head_def, head_state)
            return head(carry, x)
        carry, out = jax.vmap(eval_heads, in_axes=(0, 0, None), out_axes=(0, 0))(head_states, carry, out)

        out = self.output_norm(out)
        out = einops.rearrange(out, "heads batch hidden -> batch (heads hidden)")

        out = self.up_proj_1(out) * nnx.gelu(self.up_proj_2(out))
        out = self.down_proj(out)

        out += x  # Residual

        return carry, out

    @classmethod
    def init_carry(cls, batch_size: int, hidden_size: int, num_heads: int, rngs: nnx.Rngs):
        @nnx.split_rngs(splits=num_heads)
        @nnx.vmap(axis_size=num_heads)
        def create_carry(rngs: nnx.Rngs):
            return sLSTMCell.init_carry(batch_size, hidden_size, rngs)
        return create_carry(rngs)


class mLSTMCell(nnx.Module):
    query_proj: nnx.Linear
    key_proj: nnx.Linear
    value_proj: nnx.Linear
    qk_conv: nnx.Conv

    input_proj: nnx.Linear
    forget_proj: nnx.Linear
    output_proj: nnx.Linear

    learnable_skip: nnx.Linear

    hidden_size: int

    def __init__(self, hidden_size: int, rngs: nnx.Rngs, input_size: Optional[int] = None):
        self.hidden_size = hidden_size

        construct_qkv_proj = partial(
            nnx.Linear,
            in_features=input_size if input_size else hidden_size,
            out_features=hidden_size,
            use_bias=True,
            rngs=rngs,
        )
        construct_x_proj = partial(
            nnx.Linear,
            in_features=input_size if input_size else hidden_size,
            out_features=1,
            use_bias=True,
            rngs=rngs,
        )

        self.query_proj = construct_qkv_proj()
        self.key_proj = construct_qkv_proj()
        self.value_proj = construct_qkv_proj()
        self.qk_conv = nnx.Conv(
            in_features=1,
            out_features=1,
            kernel_size=4,
            rngs=rngs
        )

        self.input_proj = construct_x_proj(
            bias_init=nnx.initializers.normal(1e-2)
        )

        self.forget_proj = construct_x_proj(
            bias_init=uniform(3.0, 6.0)
        )

        self.output_proj = construct_x_proj(out_features=self.hidden_size)

        self.learnable_skip = nnx.Linear(
            in_features=hidden_size,
            out_features=hidden_size,
            rngs=rngs,
        )

    def __call__(self, carry, x):
        cell_state, normalizer_state, stabilizer_state = carry
        # HACK: See init_carry for an explanation
        cell_state = einops.rearrange(cell_state, "(b h1 h2) -> b h1 h2", h1=self.hidden_size, h2=self.hidden_size)
        normalizer_state = einops.rearrange(normalizer_state, "(b h) -> b h", h=self.hidden_size)

        out = nnx.sigmoid(self.output_proj(x))

        i_tilde = jnp.squeeze(self.input_proj(x), axis=1)
        f_tilde = jnp.squeeze(self.forget_proj(x), axis=1)

        # print(f"i_tilde shape: {i_tilde.shape}")
        # print(f"f_tilde shape: {f_tilde.shape}")
        # print(f"stabilizer_state shape: {stabilizer_state.shape}")

        m = jnp.maximum(f_tilde + stabilizer_state, i_tilde)  # Stabilizer state
        i = jnp.exp(i_tilde - m)
        f = jnp.exp(f_tilde + stabilizer_state - m)

        qk_input = self.qk_conv(x[..., jnp.newaxis])
        qk_input = jnp.squeeze(nnx.silu(qk_input), axis=2)

        key = self.key_proj(qk_input) / jnp.sqrt(self.hidden_size)
        query = self.query_proj(qk_input)
        value = self.value_proj(x)

        # print(f"f shape: {f.shape}")
        c = jnp.einsum("b,bij->bij", f, cell_state) + jnp.einsum("b,bi,bj->bij", i, value, key)
        n = jnp.einsum("b,bi->bi", f, normalizer_state) + jnp.einsum("b,bi->bi", i, key)

        scaler = jnp.abs(jnp.einsum("bi,bi->b", n, query))
        scaler = 1 / jnp.maximum(scaler, 1.0)
        h = out * jnp.einsum("bik,bk,b->bi", c, query, scaler)

        h += self.learnable_skip(qk_input)

        # HACK: See init_carry for an explanation
        c = einops.rearrange(c, "b h1 h2 -> (b h1 h2)")
        n = einops.rearrange(n, "b h -> (b h)")
        return (c, n, m), h

    @classmethod
    def init_carry(cls, batch_size: int, hidden_size: int, rngs: nnx.Rngs):
        c = nnx.initializers.zeros(shape=(batch_size, hidden_size, hidden_size), dtype=jnp.float16, key=rngs())
        # HACK: CoreML does not supprot tensors of rank > 5, so we squeeze the c carry to make it fit
        c = einops.rearrange(c, "b h1 h2 -> (b h1 h2)")

        n = nnx.initializers.zeros(shape=(batch_size, hidden_size), dtype=jnp.float16, key=rngs())
        n = einops.rearrange(n, "b h -> (b h)")

        m = nnx.initializers.zeros(shape=(batch_size,), dtype=jnp.float16, key=rngs())

        return c, n, m


class mLSTMBlock(nnx.Module):
    heads: List[mLSTMCell]
    input_norm: nnx.BatchNorm
    output_norm: nnx.BatchNorm

    up_proj_1: nnx.Linear
    up_proj_2: nnx.Linear
    down_proj: nnx.Linear
    head_polling: nnx.Linear

    def __init__(self, hidden_size: int, num_heads: int, rngs: nnx.Rngs):
        @nnx.split_rngs(splits=num_heads)
        @nnx.vmap(axis_size=num_heads)
        def create_heads(rngs: nnx.Rngs):
            return mLSTMCell(2 * hidden_size, rngs)
        self.heads = create_heads(rngs)

        self.input_norm = nnx.BatchNorm(hidden_size, rngs=rngs)
        self.output_norm = nnx.BatchNorm(2 * hidden_size, rngs=rngs)

        intermediate_size = hidden_size * 2
        self.up_proj_1 = nnx.Linear(
            in_features=hidden_size,
            out_features=intermediate_size,
            rngs=rngs,
        )
        self.up_proj_2 = nnx.Linear(
            in_features=hidden_size,
            out_features=intermediate_size,
            rngs=rngs,
        )
        self.down_proj = nnx.Linear(
            in_features=intermediate_size,
            out_features=hidden_size,
            rngs=rngs,
        )
        self.head_polling = nnx.Linear(
            in_features=num_heads * intermediate_size,
            out_features=intermediate_size,
            rngs=rngs,
        )

    def __call__(self, carry, x):
        out = self.input_norm(x)

        triggers = nnx.silu(self.up_proj_2(out))
        mlstm_input = self.up_proj_1(out)

        head_def, head_states = nnx.split(self.heads)

        def eval_heads(head_state, carry, x):
            head = nnx.merge(head_def, head_state)
            return head(carry, x)
        carry, mlstm_out = jax.vmap(eval_heads, in_axes=(0, 0, None), out_axes=(0, 0))(head_states, carry, mlstm_input)

        mlstm_out = self.output_norm(mlstm_out)
        mlstm_out = einops.rearrange(mlstm_out, "heads batch hidden -> batch (heads hidden)")
        mlstm_out = self.head_polling(mlstm_out)

        out = mlstm_out * triggers
        out = self.down_proj(out)

        out += x  # residual

        return carry, out

    @classmethod
    def init_carry(cls, batch_size: int, hidden_size: int, num_heads: int, rngs: nnx.Rngs):
        @nnx.split_rngs(splits=num_heads)
        @nnx.vmap(axis_size=num_heads)
        def create_carry(rngs: nnx.Rngs):
            return mLSTMCell.init_carry(batch_size, 2 * hidden_size, rngs)
        return create_carry(rngs)


class xLSTMModule(nnx.Module):
    num_mlstm: int
    num_slstm: int
    hidden_size: int
    num_heads: int

    mlstms: List[mLSTMBlock]
    slstms: List[sLSTMBlock]

    def __init__(self, hidden_size: int, num_heads: int, num_mlstm: int, num_slstm: int, rngs: nnx.Rngs):
        self.num_mlstm = num_mlstm
        self.num_slstm = num_slstm
        self.hidden_size = hidden_size
        self.num_heads = num_heads

        self.mlstms = []
        for _ in range(num_mlstm):
            self.mlstms.append(mLSTMBlock(hidden_size, num_heads, rngs=rngs))

        self.slstms = []
        for _ in range(num_slstm):
            self.slstms.append(sLSTMBlock(hidden_size, num_heads, rngs=rngs))

    def __call__(self, carry, x):
        mlstm_carry, slstm_carry = carry

        out = x

        # Call the mLSTMs
        mlstm_carries = []
        for mlstm, c in zip(self.mlstms, zip(*mlstm_carry)):
            c, out = mlstm(c, out)
            mlstm_carries.append(c)
        mlstm_carry = [jnp.stack(c, axis=0) for c in zip(*mlstm_carries)]

        # Call the sLSTMs
        slstm_carries = []
        for slstm, c in zip(self.slstms, zip(*slstm_carry)):
            c, out = slstm(c, out)
            slstm_carries.append(c)
        slstm_carry = [jnp.stack(c, axis=0) for c in zip(*slstm_carries)]

        return (mlstm_carry, slstm_carry), out

    def init_carry(self, batch_size: int, rngs: nnx.Rngs):
        @nnx.split_rngs(splits=self.num_mlstm)
        @nnx.vmap(axis_size=self.num_mlstm)
        def create_mlstm_carry(rngs: nnx.Rngs):
            return mLSTMBlock.init_carry(batch_size, self.hidden_size, self.num_heads, rngs)

        @nnx.split_rngs(splits=self.num_slstm)
        @nnx.vmap(axis_size=self.num_slstm)
        def create_slstm_carry(rngs: nnx.Rngs):
            return sLSTMBlock.init_carry(batch_size, self.hidden_size, self.num_heads, rngs)

        return (create_mlstm_carry(rngs), create_slstm_carry(rngs))


class xLSTM(nnx.Module):
    layers: List[xLSTMModule]

    def __init__(
        self,
        hidden_size: int,
        num_heads: int,
        num_layers: int,
        rngs: nnx.Rngs,
        mlstm_per_layer: int = 1,
        slstm_per_layer: int = 1
    ):
        self.layers = []
        for _ in range(num_layers):
            self.layers.append(xLSTMModule(hidden_size, num_heads, mlstm_per_layer, slstm_per_layer, rngs))

    def __call__(self, carry, x):
        out = x
        carries = []
        for layer, c in zip(self.layers, carry):
            c, out = layer(c, out)
            carries.append(c)

        return carries, out

    def init_carry(self, batch_size: int, rngs: nnx.Rngs):
        carries = []
        for layer in self.layers:
            carries.append(layer.init_carry(batch_size, rngs))
        return carries
